#include "ParticipantUDP.h"

#include "Thing.h"

#include "Arduino/ArduinoParticipant.h"
#include "EspIdf/EspIdfParticipant.h"
#include "Posix/PosixParticipant.h"
#include "Windows/WindowsParticipant.h"

#include <string.h>

namespace RoboidControl {

ParticipantUDP::ParticipantUDP(int port) : Participant("127.0.0.1", port) {
  this->remoteSite = nullptr;
  if (this->port == 0)
    this->isIsolated = true;
  Participant::registry.Add(this);
}

ParticipantUDP::ParticipantUDP(const char* ipAddress, int port, int localPort)
    : Participant("127.0.0.1", localPort) {
  if (this->port == 0)
    this->isIsolated = true;
  else
    this->remoteSite = new Participant(ipAddress, port);
  Participant::registry.Add(this);
}

static ParticipantUDP* isolatedParticipant = nullptr;

ParticipantUDP* ParticipantUDP::Isolated() {
  if (isolatedParticipant == nullptr)
    isolatedParticipant = new ParticipantUDP(0);
  return isolatedParticipant;
}

void ParticipantUDP::begin() {
  if (this->isIsolated || this->remoteSite == nullptr)
    return;

  SetupUDP(this->port, this->remoteSite->ipAddress, this->remoteSite->port);
}

void ParticipantUDP::SetupUDP(int localPort,
                              const char* remoteIpAddress,
                              int remotePort) {
#if defined(_WIN32) || defined(_WIN64)
  Windows::ParticipantUDP* thisWindows =
      static_cast<Windows::ParticipantUDP*>(this);
  thisWindows->Setup(localPort, remoteIpAddress, remotePort);
#elif defined(__unix__) || defined(__APPLE__)
  Posix::ParticipantUDP* thisPosix = static_cast<Posix::ParticipantUDP*>(this);
  thisPosix->Setup(localPort, remoteIpAddress, remotePort);
#elif defined(ARDUINO)
  Arduino::ParticipantUDP* thisArduino =
      static_cast<Arduino::ParticipantUDP*>(this);
  thisArduino->Setup();
#elif defined(IDF_VER)
  EspIdf::ParticipantUDP* thisEspIdf =
      static_cast<EspIdf::ParticipantUDP*>(this);
  thisEspIdf->Setup(localPort, remoteIpAddress, remotePort);
#endif
  this->connected = true;
}

#pragma region Update

void ParticipantUDP::Update() {
  unsigned long currentTimeMs = Thing::GetTimeMs();

  if (this->isIsolated == false) {
    if (this->connected == false)
      begin();

    if (this->publishInterval > 0 && currentTimeMs > this->nextPublishMe) {
      ParticipantMsg* msg = new ParticipantMsg(this->networkId);
      if (this->remoteSite == nullptr)
        this->Publish(msg);
      else
        this->Send(this->remoteSite, msg);
      delete msg;

      this->nextPublishMe = currentTimeMs + this->publishInterval;
    }

    this->ReceiveUDP();
  }

  UpdateMyThings();
  UpdateOtherThings();
}

void ParticipantUDP::UpdateMyThings() {
  for (Thing* thing : this->things) {
    if (thing == nullptr)  // || thing->GetParent() != nullptr)
      continue;

    if (thing->hierarchyChanged) {
      if (!(this->isIsolated || this->networkId == 0)) {
        ThingMsg* thingMsg = new ThingMsg(this->networkId, thing);
        this->Send(this->remoteSite, thingMsg);
        delete thingMsg;

        if (thing->nameChanged) {
          NameMsg* nameMsg = new NameMsg(this->networkId, thing);
          this->Send(this->remoteSite, nameMsg);
          delete nameMsg;
        }
      }
    }

    // Why don't we do recursive?
    // Because when a thing creates a thing in the update,
    // that new thing is not sent out (because of hierarchyChanged)
    // before it is updated itself: it is immediatedly updated!
    thing->Update(false);

    if (!(this->isIsolated || this->networkId == 0)) {
      if (thing->terminate) {
        DestroyMsg* destroyMsg = new DestroyMsg(this->networkId, thing);
        this->Send(this->remoteSite, destroyMsg);
        delete destroyMsg;
      } else {
        //  Send to remote site
        if (thing->nameChanged) {
          NameMsg* nameMsg = new NameMsg(this->networkId, thing);
          this->Send(this->remoteSite, nameMsg);
          delete nameMsg;
        }
        PoseMsg* poseMsg = new PoseMsg(this->networkId, thing);
        this->Send(this->remoteSite, poseMsg);
        delete poseMsg;
        BinaryMsg* binaryMsg = new BinaryMsg(this->networkId, thing);
        this->Send(this->remoteSite, binaryMsg);
        delete binaryMsg;
      }
    }
    if (thing->terminate)
      this->Remove(thing);
  }
}

void ParticipantUDP::UpdateOtherThings() {
#if defined(NO_STD)
  Participant** participants = Participant::registry.GetAll();
  for (int ix = 0; ix < Participant::registry.count; ix++) {
    Participant* participant = participants[ix];
#else
  for (Participant* participant : Participant::registry.GetAll()) {
#endif
    if (participant == nullptr || participant == this)
      continue;

    // Call only the Participant version of the Update.
    // This is to deal with the function where one of the (remote)
    // participants is actually a local participant
    participant->Participant::Update();
    if (this->isIsolated)
      continue;

    for (Thing* thing : participant->things) {
      PoseMsg* poseMsg = new PoseMsg(participant->networkId, thing);
      this->Send(participant, poseMsg);
      delete poseMsg;
      BinaryMsg* binaryMsg = new BinaryMsg(participant->networkId, thing);
      this->Send(participant, binaryMsg);
      delete binaryMsg;
    }
  }
}

// Update
#pragma endregion

#pragma region Send

void ParticipantUDP::SendThingInfo(Participant* remoteParticipant,
                                   Thing* thing) {
  // std::cout << "Send thing info [" << (int)thing->id << "] \n";
  ThingMsg* thingMsg = new ThingMsg(this->networkId, thing);
  this->Send(remoteParticipant, thingMsg);
  delete thingMsg;
  NameMsg* nameMsg = new NameMsg(this->networkId, thing);
  this->Send(remoteParticipant, nameMsg);
  delete nameMsg;
  ModelUrlMsg* modelMsg = new ModelUrlMsg(this->networkId, thing);
  this->Send(remoteParticipant, modelMsg);
  delete modelMsg;
  PoseMsg* poseMsg = new PoseMsg(this->networkId, thing, true);
  this->Send(remoteParticipant, poseMsg);
  delete poseMsg;
  BinaryMsg* customMsg = new BinaryMsg(this->networkId, thing);
  this->Send(remoteParticipant, customMsg);
  delete customMsg;
}

bool ParticipantUDP::Send(Participant* remoteParticipant, IMessage* msg) {
  int bufferSize = msg->Serialize(this->buffer);
  if (bufferSize <= 0)
    return true;

  // std::cout << "send msg " << (int)this->buffer[0] << " to "
  // << remoteParticipant->ipAddress << std::endl;

#if defined(_WIN32) || defined(_WIN64)
  Windows::ParticipantUDP* thisWindows =
      static_cast<Windows::ParticipantUDP*>(this);
  return thisWindows->Send(remoteParticipant, bufferSize);
#elif defined(__unix__) || defined(__APPLE__)
  Posix::ParticipantUDP* thisPosix = static_cast<Posix::ParticipantUDP*>(this);
  return thisPosix->Send(remoteParticipant, bufferSize);
#elif defined(ARDUINO)
  Arduino::ParticipantUDP* thisArduino =
      static_cast<Arduino::ParticipantUDP*>(this);
  return thisArduino->Send(remoteParticipant, bufferSize);
#elif defined(IDF_VER)
  EspIdf::ParticipantUDP* thisEspIdf =
      static_cast<EspIdf::ParticipantUDP*>(this);
  return thisEspIdf->Send(remoteParticipant, bufferSize);
#else
  return false;
#endif
}

void ParticipantUDP::PublishThingInfo(Thing* thing) {
  // std::cout << "Publish thing info" << thing->networkId << "\n";
  //  Strange, when publishing, the network id is irrelevant, because it is
  //  connected to a specific site...
  ThingMsg* thingMsg = new ThingMsg(this->networkId, thing);
  this->Publish(thingMsg);
  delete thingMsg;
  NameMsg* nameMsg = new NameMsg(this->networkId, thing);
  this->Publish(nameMsg);
  delete nameMsg;
  ModelUrlMsg* modelMsg = new ModelUrlMsg(this->networkId, thing);
  this->Publish(modelMsg);
  delete modelMsg;
  PoseMsg* poseMsg = new PoseMsg(this->networkId, thing, true);
  this->Publish(poseMsg);
  delete poseMsg;
  BinaryMsg* customMsg = new BinaryMsg(this->networkId, thing);
  this->Publish(customMsg);
  delete customMsg;
}

bool ParticipantUDP::Publish(IMessage* msg) {
  // std::cout << "publish msg\n";
#if defined(_WIN32) || defined(_WIN64)
  Windows::ParticipantUDP* thisWindows =
      static_cast<Windows::ParticipantUDP*>(this);
  return thisWindows->Publish(msg);
#elif defined(__unix__) || defined(__APPLE__)
  Posix::ParticipantUDP* thisPosix = static_cast<Posix::ParticipantUDP*>(this);
  return thisPosix->Publish(msg);
#elif defined(ARDUINO)
  Arduino::ParticipantUDP* thisArduino =
      static_cast<Arduino::ParticipantUDP*>(this);
  return thisArduino->Publish(msg);
#elif defined(IDF_VER)
  EspIdf::ParticipantUDP* thisEspIdf =
      static_cast<EspIdf::ParticipantUDP*>(this);
  return thisEspIdf->Publish(msg);
#else
  return false;
#endif
}

// Send
#pragma endregion

#pragma region Receive

void ParticipantUDP::ReceiveUDP() {
#if defined(_WIN32) || defined(_WIN64)
  Windows::ParticipantUDP* thisWindows =
      static_cast<Windows::ParticipantUDP*>(this);
  thisWindows->Receive();
#elif defined(__unix__) || defined(__APPLE__)
  Posix::ParticipantUDP* thisPosix = static_cast<Posix::ParticipantUDP*>(this);
  thisPosix->Receive();
#elif defined(ARDUINO)
  Arduino::ParticipantUDP* thisArduino =
      static_cast<Arduino::ParticipantUDP*>(this);
  thisArduino->Receive();
#elif defined(IDF_VER)
  EspIdf::ParticipantUDP* thisEspIdf =
      static_cast<EspIdf::ParticipantUDP*>(this);
  thisEspIdf->Receive();
#endif
}

void ParticipantUDP::ReceiveData(unsigned char packetSize,
                                 char* senderIpAddress,
                                 unsigned int senderPort) {
  // std::cout << "Receive data from " << senderIpAddress << ":" << senderPort
  //           << std::endl;
  Participant* sender = this->registry.Get(senderIpAddress, senderPort);
  if (sender == nullptr) {
    sender = this->registry.Add(senderIpAddress, senderPort);
#if !defined(NO_STD)
    std::cout << "New remote participant " << sender->ipAddress << ":"
              << sender->port << std::endl;
#endif
  }

  ReceiveData(packetSize, sender);
}

void ParticipantUDP::ReceiveData(unsigned char bufferSize,
                                 Participant* sender) {
  unsigned char msgId = this->buffer[0];
  // std::cout << "receive msg " << (int)msgId << "\n";
  // std::cout << "  buffer size = " <<(int) bufferSize << "\n";
  switch (msgId) {
    case ParticipantMsg::id: {
      ParticipantMsg* msg = new ParticipantMsg(this->buffer);
      bufferSize -= msg->length;
      Process(sender, msg);
      delete msg;
    } break;
    case NetworkIdMsg::id: {
      NetworkIdMsg* msg = new NetworkIdMsg(this->buffer);
      bufferSize -= msg->length;
      Process(sender, msg);
      delete msg;
    } break;
    case InvestigateMsg::id: {
      InvestigateMsg* msg = new InvestigateMsg(this->buffer);
      Process(sender, msg);
      delete msg;
    } break;
    case ThingMsg::id: {
      ThingMsg* msg = new ThingMsg(this->buffer);
      bufferSize -= msg->length;
      Process(sender, msg);
      delete msg;
    } break;
    case NameMsg::id: {
      NameMsg* msg = new NameMsg(this->buffer);
      bufferSize -= msg->length + msg->nameLength;
      Process(sender, msg);
      delete msg;
    } break;
    case ModelUrlMsg::id: {
      ModelUrlMsg* msg = new ModelUrlMsg(this->buffer);
      bufferSize -= msg->length + msg->urlLength;
      Process(sender, msg);
      delete msg;
    } break;
    case PoseMsg::id: {
      PoseMsg* msg = new PoseMsg(this->buffer);
      bufferSize -= msg->length;
      Process(sender, msg);
      delete msg;
    } break;
    case BinaryMsg::id: {
      BinaryMsg* msg = new BinaryMsg(this->buffer);
      bufferSize -= msg->length + msg->dataLength;
      Process(sender, msg);
      delete msg;
    } break;
  };

// Check if the buffer has been read completely
#if !defined(NO_STD)
  if (bufferSize > 0)
    std::cout << "Buffer not fully read, remaining " << (int)bufferSize << "\n";
#endif
}

void ParticipantUDP::Process(Participant* sender, ParticipantMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": Process ParticipantMsg " << (int)msg->networkId
            << "\n";
#endif
}

void ParticipantUDP::Process(Participant* sender, NetworkIdMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": process SiteMsg " << (int)this->networkId
            << " -> " << (int)msg->networkId << "\n";
#endif

  if (this->networkId != msg->networkId) {
    this->networkId = msg->networkId;

    // std::cout << this->things.size() << " things\n";
    for (Thing* thing : this->things)
      this->SendThingInfo(sender, thing);
  }
}

void ParticipantUDP::Process(Participant* sender, InvestigateMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": Process InvestigateMsg [" << (int)msg->networkId
            << "/" << (int)msg->thingId << "]\n";
#endif
}

void ParticipantUDP::Process(Participant* sender, ThingMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": process ThingMsg [" << (int)msg->networkId
            << "/" << (int)msg->thingId << "] " << (int)msg->thingType << " "
            << (int)msg->parentId << "\n";
#endif
}

void ParticipantUDP::Process(Participant* sender, NameMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": process NameMsg [" << (int)msg->networkId << "/"
            << (int)msg->thingId << "] ";
#endif

  Thing* thing = sender->Get(msg->thingId);
  if (thing != nullptr) {
    int nameLength = msg->nameLength;
    int stringLen = nameLength + 1;
    char* thingName = new char[stringLen];
#if defined(_WIN32) || defined(_WIN64)
    strncpy_s(thingName, stringLen, msg->name,
              stringLen - 1);  // Leave space for null terminator
#else
    // Use strncpy with bounds checking for other platforms (Arduino, POSIX,
    // ESP-IDF)
    strncpy(thingName, msg->name,
            nameLength);  // Leave space for null terminator
#endif
    thingName[nameLength] = '\0';
    thing->SetName(thingName);

#if !defined(NO_STD)
    std::cout << thing->GetName();
#endif
  }
#if !defined(NO_STD)
  std::cout << std::endl;
#endif
}

void ParticipantUDP::Process(Participant* sender, ModelUrlMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": process ModelUrlMsg [" << (int)msg->networkId
            << "/" << (int)msg->thingId << "]\n";
#endif
}

void ParticipantUDP::Process(Participant* sender, PoseMsg* msg) {
#if !defined(DEBUG) && !defined(NO_STD)
  std::cout << this->name << ": process PoseMsg [" << (int)this->networkId
            << "/" << (int)msg->networkId << "] " << (int)msg->poseType << "\n";
#endif
  Participant* owner = Participant::registry.Get(msg->networkId);
  if (owner == nullptr)
    return;

  Thing* thing = owner->Get(msg->thingId);
  if (thing == nullptr)
    return;

  if ((msg->poseType & PoseMsg::Pose_Position) != 0)
    thing->SetPosition(msg->position);
  if ((msg->poseType & PoseMsg::Pose_Orientation) != 0)
    thing->SetOrientation(msg->orientation);
  if ((msg->poseType & PoseMsg::Pose_LinearVelocity) != 0)
    thing->SetLinearVelocity(msg->linearVelocity);
  if ((msg->poseType & PoseMsg::Pose_AngularVelocity) != 0)
    thing->SetAngularVelocity(msg->angularVelocity);
}

void ParticipantUDP::Process(Participant* sender, BinaryMsg* msg) {
#if defined(DEBUG)
  std::cout << this->name << ": process BinaryMsg [" << (int)msg->networkId
            << "/" << (int)msg->thingId << "]\n";
#endif

  Participant* owner = Participant::registry.Get(msg->networkId);
  if (owner != nullptr) {
    Thing* thing = owner->Get(msg->thingId);
    if (thing != nullptr)
      thing->ProcessBinary(msg->data);
#if !defined(NO_STD)
    else {
#if defined(DEBUG)
      std::cout << "  unknown thing [" << (int)msg->networkId << "/"
                << (int)msg->thingId << "]";
#endif
    }
#endif
  }
}

// Receive
#pragma endregion

}  // namespace RoboidControl
